SKIN = {
    NAME    = "Aurora",
    AUTHOR  = "Naughty_Cat",
    VERSION = "1.3.8",
}

options = {
    allow_puzzle_ship = true, -- replaces a vehicle if current mode is puzzle (pusher, pointman, erazer, vegas, etc.)
    aurora = {
        dynamic = {
            enable = true,
            color_source = "highway", -- "highway", "highwayinverted"
            color_scaler = {min = 0.5, max = 1.75},
        },
    },
    random_track_colors = {
        method = "all", -- "only", "include", "exclude", "all"
        palletes = {}, -- avalaible palletes: 1, 2, 3, 4, 5
    },
    rings = {
        style = "tiny", -- "none", "standart", "double", "tiny"
        allow_glow = true,
    },
}



IS_EXPERIMENTAL = false



local skin_vars = GetSkinProperties()
local track_width = skin_vars.trackwidth
local is_mono, is_puzzle, is_wakeboard = skin_vars.colorcount == 1, skin_vars.colorcount > 1, PlayerCanJump()
-- local track = GetTrack()

local GetQualityLevel = GetQualityLevel4

local function set_by_quality(low, med, high, ultra)
    local ql = {low, med, high, ultra}
    return ql[GetQualityLevel()]
end

local function set_by_lowhigh(low, high)
    if GetQualityLevel() > 2 then return high else return low end
end

--[[
-- I'm so angry by that in-game reverse print
local print_direct = print -- immediately
do
    local print_list = {}
    
    print = function(str)
        table.insert(print_list, str)
        Profile(string.format("\nPRINT CALL (id = %i): '%s'.\n", #print_list, tostring(str)))
    end
    
    function print_all()
        for i = #print_list, 1, -1 do
            print_direct(print_list[i])
        end
    end
end
]]

local function skybox_from_dir(o)
    local t = {}
    local keys = {"_FrontTex", "_BackTex", "_LeftTex", "_RightTex", "_UpTex", "_DownTex"}
    local types = {
        full = {"front", "back", "left", "right", "top", "bottom"},
        simple = {"side", "side", "side", "side", "updown", "updown"},
    }
    for i, key in ipairs(keys) do
        t[key] = tostring(o.path) .. "/" .. tostring(types[o.type][i]) .. "." .. tostring(o.extension)
    end
    return t
end

local function arithmetic_mean(...)
    local s = 0
    for _, v in ipairs(arg) do s = s + v end
    return (s / #arg)
end

local function one_value_table(v, r)
    local t = {}
    for i = 1, r do table.insert(t, v) end
    return t
end

local function min_max_value(v, min, max)
    if v < min then
        return min
    elseif v > max then
        return max
    else
        return v
    end
end

-- a simple array shuffle function for Lua implementing the Fisher-Yates shuffle
-- https://gist.github.com/Uradamus/10323382
local function table_shuffle(tbl)
    local size = #tbl
    for i = size, 1, -1 do
        local rand = math.random(size)
        tbl[i], tbl[rand] = tbl[rand], tbl[i]
    end
    return tbl
end

local function table_merge(t1, t2)
    for k,v in pairs(t2) do
        if type(v) == "table" then
            if type(t1[k] or false) == "table" then
                table_merge(t1[k] or {}, t2[k] or {})
            else
                t1[k] = v
            end
        else
            t1[k] = v
        end
    end
    return t1
end

-- counts all keys in a table, not just numeric keys like '#'
local function table_count(tbl)
    local c = 0
    for _, _ in pairs(tbl) do c = c + 1 end
    return c
end

local function switch(cond, cases)
    return fif(cases[cond], cases[cond], cases["default"])
end

-- https://stackoverflow.com/a/30757399
local function is_table_equal(t1, t2, ignore_mt)
    local ty1 = type(t1)
    local ty2 = type(t2)
    if ty1 ~= ty2 then return false end
    -- non-table types can be directly compared
    if ty1 ~= 'table' and ty2 ~= 'table' then return t1 == t2 end
    -- as well as tables which have the metamethod __eq
    local mt = getmetatable(t1)
    if not ignore_mt and mt and mt.__eq then return t1 == t2 end
    for k1,v1 in pairs(t1) do
        local v2 = t2[k1]
        if v2 == nil or not is_table_equal(v1,v2) then return false end
    end
    for k2,v2 in pairs(t2) do
        local v1 = t1[k2]
        if v1 == nil or not is_table_equal(v1,v2) then return false end
    end
    return true
end

local function super_switch(cond, cases)
    local default = cases["default"]
    local case = fif(cases[cond], cases[cond], default)
    if is_table_equal(case, default) then return default end
    for k, v in pairs(default) do
        if case[k] == nil then case[k] = deepcopy(default[k]) end
    end
    return case
end

local function is_shoulders_exist()
    if not skin_vars.shoulderlines then return false end
    if #skin_vars.shoulderlines == 0 then return false end

    local left_shoulder = math.floor(-track_width) ~= math.ceil(skin_vars.shoulderlines[1])
    local right_shoulder = math.ceil(skin_vars.shoulderlines[2]) ~= math.floor(track_width)

    return left_shoulder and right_shoulder;
end

local function selective_table_random(tbl, param, param_values)
    local params = {
        only = function() -- "param_values" must be the string (or isn't?)
            return tbl[param_values[1]]
        end,
        include = function()
            -- print("included: "..dump_table(param_values))
            local randomed_value = math.random(table_count(param_values))
            return tbl[param_values[randomed_value]]
        end,
        exclude = function()
            local param_values_e = deepcopy(param_values) -- not excluded
            param_values = {}
            for key, _ in pairs(tbl) do
                local is_excluded = false
                for _, excluded in ipairs(param_values_e) do
                    if not is_excluded then is_excluded = key == excluded end
                end
                if not is_excluded then table.insert(param_values, key) end
            end
            -- print("not excluded: "..dump_table(param_values))
            local randomed_value = math.random(table_count(param_values))
            return tbl[param_values[randomed_value]]
        end,
        all = function()
            local c, key, key_pos = 1, "", math.random(table_count(tbl))
            for k, v in pairs(tbl) do
                if c == key_pos then key = k end
                c = c + 1
            end
            return tbl[key]
        end,
    }

    return params[param]()
end

-- best function for dumping a table
-- https://stackoverflow.com/a/42062321
local function dump_table(node)
    -- to make output beautiful
    local function tab(amt)
        local str = ""
        for i=1,amt do
            str = str .. "\t"
        end
        return str
    end

    local cache, stack, output = {},{},{}
    local depth = 1
    local output_str = "{\n"

    while true do
        local size = 0
        for k,v in pairs(node) do
            size = size + 1
        end

        local cur_index = 1
        for k,v in pairs(node) do
            if (cache[node] == nil) or (cur_index >= cache[node]) then

                if (string.find(output_str,"}",output_str:len())) then
                    output_str = output_str .. ",\n"
                elseif not (string.find(output_str,"\n",output_str:len())) then
                    output_str = output_str .. "\n"
                end

                -- This is necessary for working with HUGE tables otherwise we run out of memory using concat on huge strings
                table.insert(output,output_str)
                output_str = ""

                local key
                if (type(k) == "number" or type(k) == "boolean") then
                    key = "["..tostring(k).."]"
                else
                    key = "['"..tostring(k).."']"
                end

                if (type(v) == "number" or type(v) == "boolean") then
                    output_str = output_str .. tab(depth) .. key .. " = "..tostring(v)
                elseif (type(v) == "table") then
                    output_str = output_str .. tab(depth) .. key .. " = {\n"
                    table.insert(stack,node)
                    table.insert(stack,v)
                    cache[node] = cur_index+1
                    break
                else
                    output_str = output_str .. tab(depth) .. key .. " = '"..tostring(v).."'"
                end

                if (cur_index == size) then
                    output_str = output_str .. "\n" .. tab(depth-1) .. "}"
                else
                    output_str = output_str .. ","
                end
            else
                -- close the table
                if (cur_index == size) then
                    output_str = output_str .. "\n" .. tab(depth-1) .. "}"
                end
            end

            cur_index = cur_index + 1
        end

        if (size == 0) then
            output_str = output_str .. "\n" .. tab(depth-1) .. "}"
        end

        if (#stack > 0) then
            node = stack[#stack]
            stack[#stack] = nil
            depth = cache[node] == nil and depth + 1 or depth - 1
        else
            break
        end
    end

    -- This is necessary for working with HUGE tables otherwise we run out of memory using concat on huge strings
    table.insert(output,output_str)
    output_str = table.concat(output)

    return output_str
end

do
    local tt_keys_list = {}

    local function value(keys, tbl, is_return)
        local is_return = is_return or false
    
        -- get existed keys
        local e_k = {} 
        for i, v in ipairs(keys) do
            if not tbl[v] then
                table.insert(e_k, v)
            end
        end
        
        for i, v in ipairs(e_k) do
            tbl[v] = tbl[i]
            tbl[i] = nil
        end
        
        -- remove unnecessary keys
        local is_necessary = false
        for t_k, v in pairs(tbl) do
            for i, r_k in ipairs(keys) do
                if t_k == r_k then
                    is_necessary = true
                    break
                end
            end
            if not is_necessary then
                tbl[t_k] = nil
            end
            is_necessary = false
        end
        
        if is_return then return tbl end
    end
    
    local function table_of_values(keys, tbl_of_tbl, is_return)
        local is_return = is_return or false
        
        for i = 1, #tbl_of_tbl do
            value(keys, tbl_of_tbl[i])
        end
        
        if is_return then return tbl_of_tbl end
    end

    function create_table_type(...)
        local keys = deepcopy(arg); keys.n = nil;

        local tt, t_o_tt = {}, {}
        
        tt.__keys = deepcopy(keys)
        t_o_tt.__keys = deepcopy(keys)

        setmetatable(tt, {
            __call = function(self, o)
                value(self.__keys, o, false)
            end
        })
        
        tt.new = function(self, o)
            return value(self.__keys, o, true)
        end
        
        setmetatable(t_o_tt, {
            __call = function(self, o)
                table_of_values(self.__keys, o, false)
            end
        })

        t_o_tt.new = function(self, o)
            return table_of_values(self.__keys, o, true)
        end

        return tt, t_o_tt
    end
end

local create_table_type = create_table_type

local rgb, table_of_rgb = create_table_type("r", "g", "b")
-- local _rgba, _table_of_rgba = create_table_type("r", "g", "b", "a")
-- local _vec2, _table_of_vec2 = create_table_type("x", "y")
-- local _vec3, _table_of_vec3 = create_table_type("x", "y", "z")

--[[
local test_rgb = {{255, 0, 0}, {255, 255, 0}}
table_of_rgb(test_rgb)
print(dump_table(test_rgb))

local test_vec2 = {{255, 0, 0}, {255, 255, 0}}
_table_of_vec2(test_vec2)
print(dump_table(test_vec2))
]]

--[[ maybe in next time, boy
local function create_onupdate_listener_handler()
    local update_list = {}

    local tl_r = -1
    local p_tl_r = -1
    local p_ti = -1

    return function(params)
        table.insert(update_list, {
            listener = params.listener, 
            time = 0, 
            rate = params.rate or 0.001,
            update_type = params.update_type or "normal", -- "track_location", "intensity"
            update_on_pause = fif(params.update_on_pause ~= nil, params.update_on_pause, false),
        })
    end,
    function(dt, tl, ps, pjh, ti) -- dt, tl, ps, pjh, ti  (dt, trackLocation, playerStrafe, playerJumpHeight, intensity)
        tl_r = math.floor(tl)
        for i, listener in ipairs(update_list) do
            if dt ~= 0 and (not listener.update_on_pause) then
                local u_t_conds = switch(listener.update_type, {
                    normal = true,
                    track_location = tl_r > p_tl_r,
                    intensity = ti ~= p_ti,
                })

                if u_t_conds then
                    listener.time = listener.time + dt
                    if listener.time > listener.rate then
                        listener.time = listener.time - listener.rate
                        listener.listener(dt, tl, ps, pjh, ti)
                    end
                end
            end
        end

        p_tl_r = tl_r
        p_ti = ti
    end
end
]]

local function create_onupdate_listener_handler()
    local update_list = {}
    local listener_curr_id

    function set_delay(sec)
        update_list[listener_curr_id].delay_rate = sec
    end
    
    return function(listener, rate)
        table.insert(update_list, {listener = listener, rate = rate or 0.001, time = 0, delay = 0, delay_rate = 0})
    end,
    function(dt, tl, ps, pjh, ti) -- dt, trackLocation, playerStrafe, playerJumpHeight, intensity
        if dt == 0 then return end

        for i, listener in ipairs(update_list) do
            listener_curr_id = i
            if listener.delay_rate == 0 then
                listener.time = listener.time + dt
                if listener.time > listener.rate then
                    listener.time = listener.time - listener.rate
                    listener.listener(dt, tl, ps, pjh, ti)
                end
            else
                listener.delay = listener.delay + dt
                if listener.delay > listener.delay_rate then 
                    listener.delay = 0
                    listener.delay_rate = 0
                end
            end
        end
    end
end
-- local set_delay = set_delay -- it doesn't work

local add_update_listener, onupdate = create_onupdate_listener_handler()

-- getting a number value from a percent value in the number range
local function value_from_range(percent, first, last)
    return ((last - first) * percent) + first
end

-- getting a percent value from a number value in the number range
local function percent_from_range(value, first, last)
    return (value - first) / (last - first)
end

--[[
    Audiosurf 2 supports two methods to describe RGB colors. 
    First method is use a float numbers between 0 and 1 {0.5, 0.25, 0.15}.
    Second method is use a byte numbers {178,  34,  34}.
    I prefer to use the second method, but one very annoying thing appears because of that. 
    And this thing is when I'm trying to describe this color: {1, 1, 1}.
    For the first method it is correct, it's the white color.
    But for the second method when this color supposes to be the almost black color, it just throws away because of the first method.

    For this annoying thing i made some wrapper function, which is just convert the second method to the first method. (rgb_normalize)
]]

local function rgb_set_color_by_values(rgb_tbl, r, g, b)
    rgb_tbl.r = r; rgb_tbl.g = g; rgb_tbl.b = b;
end

local function rgb_set_color_by_table(rgb_tbl, other_rgb_tbl)
    rgb_tbl.r = other_rgb_tbl.r; rgb_tbl.g = other_rgb_tbl.g; rgb_tbl.b = other_rgb_tbl.b;
end

local function rgb_is_equal(rgb1, rgb2)
    return rgb1.r == rgb2.r and rgb1.g == rgb2.g and rgb1.b == rgb2.b
end
    
local function rgb_normalize_color(rgb_tbl)
    for k, v in pairs(rgb_tbl) do
        rgb_tbl[k] = v / 255
    end
end

local function rgb_restore_color(rgb_tbl)
    for k, v in pairs(rgb_tbl) do
        rgb_tbl[k] = math.floor(v * 255)
    end
end

local function rgb_inverse_color(rgb_tbl)
    for k, v in pairs(rgb_tbl) do
        rgb_tbl[k] = 255 - v
    end
end

local function rgb_get_color_brightness(rgb_tbl)
    rgb_normalize_color(rgb_tbl)
    return arithmetic_mean(rgb_tbl.r, rgb_tbl.g, rgb_tbl.b)
end

-- Try to make the colorscaler
-- This is the almost same colorscaler from the game
local function rgb_scale_color(rgb_tbl, scale)
    for k, v in pairs(rgb_tbl) do
        rgb_tbl[k] = math.floor(min_max_value(v * scale, 0, 255))
    end
end

-- Other implementarion. (from very dark (0) to very bright (2), '1' returns the color which is passed)
local function rgb_bright_color(rgb_tbl, brightness)
    if brightness > 1 then
        for k, v in pairs(rgb_tbl) do
            rgb_tbl[k] = math.floor(min_max_value(((255 - v) * (brightness - 1)) + v, 0, 255))
        end
    elseif brightness < 1 then
        for k, v in pairs(rgb_tbl) do
            rgb_tbl[k] = math.floor(min_max_value(v * brightness, 0, 255))
        end
    end
end

local function rgb_to_string(rgb_tbl)
    return string.format("{r = %d, g = %d, b = %d}", rgb_tbl.r, rgb_tbl.g, rgb_tbl.b)
end

local function rgb_table_to_string(tbl_of_rgb_tbl)
    local rgb_tbl_str = "{\n\t"
    for i, color in ipairs(tbl_of_rgb_tbl) do
        rgb_tbl_str = rgb_tbl_str .. string.format("[%d] = ", i) .. rgb_to_string(color) .. fif(i < #tbl_of_rgb_tbl, ",\n\t", "")
    end
    rgb_tbl_str = rgb_tbl_str .. "\n}"
    return rgb_tbl_str
end

local function build_material_and_get_resources(src)
    local update_src_data = {}
    if src.update_init then
        update_src_data = deepcopy(src.update_init)
        src.update_init = nil
    end
    mtr = BuildMaterial(src)
    update_src_data.material = mtr
    return mtr, fif(table_count(update_src_data), update_src_data, nil)
end

local function strict_value(src, rel_type, default_value)
    local is_filled_table = true
    if rel_type == "table" then
        is_filled_table = table_count(src) > 0
    end
    if src and type(src) == rel_type and is_filled_table then
        return src
    else
        if default_value then
            return default_value
        else
            if related_type == "number" then return 0 end
            if related_type == "string" then return "" end
            -- if related_type == "boolean" then return false end
        end
    end
end

-- try to make the simular colorscaler handler
do
    local dt, trackLocation, playerStrafe, playerJumpHeight, intensity = 0, 0, 0, 0, 0 -- arguments as upvalues
    
    local mtr_list = {}
    local highway_color = rgb:new {0, 0, 0}

    local function get_colorsource_handler(colorsource_param)
        if type(colorsource_param) == "string" then
            return switch(colorsource_param, {

                static = function()
                    return {r = 255, g = 255, b = 255}
                end,
                
                highway = function()
                    return highway_color
                end,

                highwayinverted = function()
                    rgb_inverse_color(highway_color)
                    return highway_color
                end,

            })
        else
            return function()
                return highway_color
            end
        end 
    end

    local function get_scaletype_handler(scaletype_param)
        if type(scaletype_param) == "string" then
            return switch(scaletype_param, {

                intensity = function()
                    return intensity -- GetIntensity()
                end,

            })
        else
            return function()
                return intensity -- GetIntensity()
            end
        end 
    end

    function material_onupdate_attach_colorscaler_handler(props)
        table.insert(mtr_list, {
            update_data = props.material_update_data,
            colorsource = get_colorsource_handler(props.colorsource),
            scaletype = get_scaletype_handler(props.scaletype),
            minscaler = props.minscaler,
            maxscaler = props.maxscaler,
        })
    end

    function colorscaler_handler(dt, tl, ps, pjh, ti) -- the color brightness adjustment variation
        if dt == 0 then return end

        dt, trackLocation, playerStrafe, playerJumpHeight, intensity = dt, tl, ps, pjh, ti -- arguments as upvalues

        rgb_set_color_by_table(highway_color, GetHighwayColor())
        rgb_restore_color(highway_color)

        for i, mtr in ipairs(mtr_list) do
            rgb_set_color_by_table(mtr.update_data.shadercolors._Color, mtr.colorsource())

            -- print_direct("original color: " .. rgb_to_string(mtr.update_data.shadercolors._Color))

            rgb_scale_color(mtr.update_data.shadercolors._Color, value_from_range(mtr.scaletype(), mtr.minscaler, mtr.maxscaler))

            -- print_direct("scaled color: " .. rgb_to_string(mtr.update_data.shadercolors._Color))

            UpdateMaterial(mtr.update_data)
        end
    end

    add_update_listener(colorscaler_handler, 0.025)
end



SetScene{
    glowpasses = 2,
    glowspread = 2,

    radialblur_strength = 0.5,

    water = is_wakeboard,
    towropes = is_wakeboard,
    watertype = 2,
    watertint = {11, 11, 11, 11},
    watertexture = "wtr_waterBW.png",
    minwaterrandomdarkness = 10, 
    numvisiblewaterchunks = set_by_lowhigh(2, 4),
    
    airdebris_count = set_by_lowhigh(5000, 8500),
    airdebris_density = set_by_lowhigh(55, 80),
    airdebris_texture = "scn_star_alpha.png",
    
    hide_default_background_verticals = true,
    useblackgrid = false
}

SetBlockFlashes{sizescaler_missed = 0.1, sizescaler = 0.5}

LoadSounds{
    hit = "snd_collision_color.wav",
    hitgrey = "snd_collision_stone.wav",
    hitgreypro = "snd_collision_stone.wav",
    matchsmall = "snd_matchsmall.wav",
    matchmedium = "snd_matchmedium.wav",
    matchlarge = "snd_matchlarge.wav",
    matchhuge = "snd_matchhuge.wav",
    landcrash = "snd_collision_black.wav"
}

if is_wakeboard then
    SetPlayer{
        cameramode = "first_jumpthird",
        camfirst = {
            pos = {0, 2.7, -3.50475},
            rot = {20.49113, 0, 0},
            strafefactor = 1
        },
        camthird = {
            pos  = {0, 2.7, -3.50475},
            pos2 = {0, 2.8, -3.50475},
            rot  = {20.49113, 0, 0},
            rot2 = {20.49113, 0, 0},
            strafefactor = 0.75,
            strafefactorFar = 1
        },
        surfer = {
            use_water_rooster = true,
            roostercolor = {25, 25, 25},
            arms = {
                shader = "VertexColorUnlitTinted",
                shadercolors = {_Color = {colorsource = "highway", scaletype = "intensity", minscaler = 1.75, maxscaler = 2.25}},
                texture = "pms_FullLeftArm_1024_wAO.png"
            },
            board = {
                shader = "VertexColorUnlitTinted",
                shadercolors = {_Color = {colorsource = "highway", scaletype = "intensity", minscaler = 1.75, maxscaler = 2.25}},
                texture = "pms_board_internalOutline.png"
            },
            body = {
                shader = "VertexColorUnlitTinted",
                shadercolors = {_Color = {colorsource = "highway", scaletype = "intensity", minscaler = 1.75, maxscaler = 2.25}},
                texture = "pms_robot_HighContrast.png"
            }
        }
    }
else
    local t = { -- deep_thrusters
        shape = {
            {-.35, -.35, 0},
            { -.5,    0, 0},
            {-.35,  .35, 0},
            {   0,   .5, 0},
            { .35,  .35, 0},
            {  .5,    0, 0},
            { .35, -.35, 0},
            {   0,  -.5, 0},
        },
        scalers = one_value_table(1, 8),
        colorscaler = 0.5,
        -- deepcount = set_by_lowhigh(1, 3)
    }

    SetPlayer{
        cameramode = "third",
        cameradynamics = "high",
        camthird = {
            pos  = {0, 2.5 , -1},
            pos2 = {0, 5.25, -4},
            rot  = {30, 0, 0},
            rot2 = {30, 0, 0},
            strafefactor    = 0.65,
            strafefactorFar = 0.85,
            puzzleoffset  = -0.95,
            puzzleoffset2 = -1.65,
            transitionspeed = 3
        },
        vehicle = fif(is_puzzle and options.allow_puzzle_ship, 
        { -- puzzle ship
            mesh = "pmv_vehicle1a.obj",
            pos = {x = 0, y = 0, z = 0.25},
            scale = {x = 1, y = 1, z = 1},
            shader = "VertexColorUnlitTintedAddSmooth",
            shadercolors = {_Color = "highway"},
            texture = "pmv_vehicle1a_ao_inverted.png",
            layer = set_by_lowhigh(14, 15),
            renderqueue = 2900,

            min_hover_height = 0.23,
            max_hover_height = 0.23, -- 0.75
            use_water_rooster = false,
            smooth_tilting = false,
            reflect = GetQualityLevel() ~= 1,
            thrusters = {
                crossSectionShape = t.shape,
                perShapeNodeColorScalers = t.scalers,
                colorscaler = t.colorscaler,
                shader = "VertexColorUnlitTintedAddSmooth",
                layer = set_by_lowhigh(14, 15),
                renderqueue = 2901,
                extrusions = 22,
                stretch = -0.1191,
                updateseconds = 0.025,
                instances = {
                    {pos = { 0.0000,  0.2810, -1.3300}, rot = {0, 0, 0}, scale = {0.7, 0.7, 0.7}},
                    {pos = { 0.1236,  0.2000, -1.2970}, rot = {0, 0, 0}, scale = {0.5, 0.5, 0.5}},
                    {pos = {-0.1236,  0.2000, -1.2970}, rot = {0, 0, 0}, scale = {0.5, 0.5, 0.5}},
                },
            }
        }, 
        { -- default ship
            mesh = "pmv_ninjamono.obj",
            pos = {x = 0, y = 0, z = 0},
            scale = {x = 1, y = 1, z = 1},
            shader = "VertexColorUnlitTintedAddSmooth",
            shadercolors = {_Color = "highway"},
            texture = "pmv_ninjamono_inverted.png",
            layer = set_by_lowhigh(14, 15),
            renderqueue = 2900,

            min_hover_height = 0.23,
            max_hover_height = 0.8,
            use_water_rooster = false,
            smooth_tilting = false,
            reflect = GetQualityLevel() ~= 1,
            thrusters = {
                crossSectionShape = t.shape,
                perShapeNodeColorScalers = t.scalers,
                colorscaler = t.colorscaler,
                shader = "VertexColorUnlitTintedAddSmooth",
                layer = set_by_lowhigh(14, 15),
                renderqueue = 2901,
                extrusions = 22,
                stretch = -0.1191,
                updateseconds = 0.025,
                instances = {
                    {pos = { 0.0000,  0.4600, -1.2800}, rot = {0, 0, 0}, scale = {0.7, 0.7, 0.7}},
                    {pos = { 0.1750,  0.2100, -1.2970}, rot = {0, 0, 0}, scale = {0.7, 0.7, 0.7}},
                    {pos = {-0.1750,  0.2100, -1.2970}, rot = {0, 0, 0}, scale = {0.7, 0.7, 0.7}},
                },
            }
        })
    }
end

if not is_wakeboard then
    SetBlocks{
        maxvisiblecount = set_by_lowhigh(50, 150),
        colorblocks = {
            float_on_water = false,

            mesh = "blo_Lozenge_fixed_HQ.obj",
            height = 0,
            scale = {1, 1, 1},
            shader = "UnlitEdgedBlock",
            texture = "blo_Lozenge_fixed_ueb.png",
            reflect = GetQualityLevel() ~= 1
        },
        greyblocks = {
            mesh = "blo_spike.obj",
            shader = "VertexColorUnlitTinted",
            texture = "blo_spike.png",
            reflect = GetQualityLevel() ~= 1
        },
        powerups = {
            powerpellet = {
                mesh = "blo_Lozenge_fixed_HQ.obj",
                height = 1.5,
                scale = {1.5, 2, 2},
                shader = "UnlitEdgedBlock",
                shadercolors = {_Color = "highwayinverted"},
                texture = "blo_Lozenge_fixed_ueb_reversed.png",
                reflect = GetQualityLevel() ~= 1
            },
            whiteblock = {
                mesh = "blo_Lozenge_fixed_HQ.obj",
                shader = "VertexColorUnlitTinted",
                texture = "blo_Lozenge_fixed_reversed.png",
                reflect = GetQualityLevel() ~= 1
            }
        }   
    }
end

SetTrackColors(selective_table_random({
    {
        {255,   0, 255}, -- Magenta
        {139,   0, 139}, -- DarkMagenta
        { 25,  25, 112}, -- Midnight Blue
        {  0, 191, 255}, -- DeepSkyBlue
        {  0, 206, 209}, -- DarkTurquoise
        {  0, 255, 255}, -- Aqua
    },
    {
        { 34, 139,  34}, -- ForestGreen
        { 60, 179, 113}, -- MediumSeaGreen
        {  0, 191, 255}, -- DeepSkyBlue
        { 30, 144, 255}, -- DodgerBlue
        {220,  20,  60}, -- Crimson
        {178,  34,  34}, -- FireBrick
    },
    {
        {  0,   0, 139}, -- DarkBlue
        { 25,  25, 112}, -- Midnight Blue
        {  0, 191, 255}, -- DeepSkyBlue
        {220,  20,  60}, -- Crimson
        {255,   0,   0}, -- Red
        {255,  69,   0}, -- OrangeRed
    },
    {
        {221, 160, 221}, -- Plum
        {216, 191, 216}, -- Thistle
        {135, 206, 235}, -- SkyBlue
        {135, 206, 250}, -- LightSkyBlue
        {240, 248, 255}, -- AliceBlue   
        {240, 255, 255}, -- Azure
    },
    {
        {128,   0, 128}, -- Purple
        {147, 112, 219}, -- MediumPurple
        { 34, 139,  34}, -- ForestGreen
        { 50, 205,  50}, -- LimeGreen
        {  0, 250, 154}, -- MediumSpringGreen
        {  0, 250, 127}, -- SpringGreen
    },
    -- {
    --     {143, 188, 143}, -- DarkSeaGreen
    --     { 85, 107,  47}, -- DarkOliveGreen
    --     {107, 142,  35}, -- OliveDrab
    --     {128, 128,   0}, -- Olive
    --     {102, 205, 170}, -- MediumAquamarine
    --     {127, 255, 212}, -- Aquamarine
    -- },
    -- {
    --     {238, 232, 170}, -- PaleGoldenrod
    --     {189, 183, 107}, -- DarkKhaki
    --     {240, 230, 140}, -- Khaki
    --     {230, 230, 250}, -- Lavender
    --     {216, 191, 216}, -- Thistle
    --     {221, 160, 221}, -- Plum
    -- },
}, options.random_track_colors.method, options.random_track_colors.palletes))

SetBlockColors((function()
    local block_colors = {
        {139,   0, 139}, -- DarkMagenta
        { 30, 144, 255}, -- DodgerBlue
        { 34, 139,  34}, -- ForestGreen
        {255, 215,   0}, -- Gold
        {199,   0,   0}, -- Red
    }

    if skin_vars.colorcount < 4 then
        table.remove(block_colors, 1)
        table.remove(block_colors, 2)
    elseif skin_vars.colorcount < 5 then
        table.remove(block_colors, 3)
    end

    return block_colors
end)())

SetPuzzleGraphics{
    puzzlematerial = {
        texture = "puz_flat3x1_tiles.png", 
        shader = "VertexColorUnlitTintedAlpha2",
        shadercolors = {_Color = {255, 255, 255, 255}}
    },
    puzzlematchmaterial = {
        texture = "puz_flat3x1_tilematch.png", 
        shader = "VertexColorUnlitTintedAlpha2",
        shadercolors = {_Color = {255, 255, 255, 255}}
    },
    puzzleflyupmaterial = {
        texture = "puz_flat3x1_tileflyup.png",
        shader = "VertexColorUnlitTintedAddFlyup"
    },
    usesublayerclone = false
}

SetSkybox{ 
    -- Due to the fact that skysphere was ruined by glow and vertical lines, I had to use skybox.
    -- And i used those sites for fast converting.
    -- https://jaxry.github.io/panorama-to-cubemap/
    -- https://www.360toolkit.co/convert-spherical-equirectangular-to-cubemap.html
    custom = skybox_from_dir{path = "sky_deep_blue_night", type = "simple", extension = "png"},
    color = {0.5, 0.5, 0.5},
}

SetRings((function()
    local rings = super_switch(options.rings.style, {
        none = {
            size = 0,
            shader = "VertexColorUnlitTintedAlpha", 
            texture = "<ALBUMART>",
        },
        standart = {
            shader = "VertexColorUnlitTintedAddSmooth", 
        },
        double = {
            shader = "VertexColorUnlitTintedAddDouble", 
        },
        tiny = {
            shader = "VertexColorUnlitTintedAddSmoothQuarter", 
        },
        default = {
            size = track_width * 2.25,
            shader = "VertexColorUnlitTintedAddSmooth", 
            texture = "rng_moonphase_smooth_alpha.png", 
        },
    })

    return {
        size = rings.size,
        offset = {0, rings.size * 0, 0},
        shader = rings.shader,
        texture = rings.texture,
        layer = fif(options.rings.allow_glow, set_by_lowhigh(13, 11), 13),
        percentringed = 0.2,
        useairrings = false,
        airsize = 0,
        airshader = "VertexColorUnlitTintedAlpha",
        airtexture = "<ALBUMART>"
    }
end)())

SetWake{
    height = fif(is_wakeboard, 2, 0),
    fallrate = .999,
    shader = "VertexColorUnlitTintedAddSmooth",
    layer = 13,
    bottomcolor = "highway",
    topcolor = {0, 0, 0}
}

CreateObject{ -- moon on the end of track
    name = "moon",
    tracknode = "end",
    gameobject = {
        mesh = "env_moon.obj",
        shader = "VertexColorUnlitTinted",
        texture = "env_moon.png",
        transform = {
            pos = {0, 15, 25},
            scale = {1.1, 1.1, 1.1}
        }
    }
}

function rails_create_border_aurora_lines(props)
    local default = {
        pos = {-track_width, track_width},
        height = 0.025,
        main_line = {
            width = 0.2,
            color = {255, 255, 255},
            color_mode = "aurora",
            material = BuildMaterial{
                shader = "VertexColorUnlitTintedQuadruple",
                layer = 13,
                renderqueue = 2750,
            },
        },
        border_line = {
            width = 0.03,
            plasement = "inner",
            color = {240, 240, 240},
            color_mode = "highway",
            material = BuildMaterial{
                shader = "VertexColorUnlitTinted",
                layer = 13,
                renderqueue = 2750,
            },
        },
    }

    local params = table_merge(default, props)
    local half_size = params.main_line.width / 2
    local is_outter = params.border_line.plasement == "outter"
    local es = params.border_line.width
    local hs = half_size
    local edges_shapes = { 
        {
            {-hs + fif(is_outter, -es, 0 ), 0}, 
            {-hs + fif(is_outter,  0 , es), 0},
        }, 
        {
            { hs + fif(is_outter, 0 , -es), 0}, 
            { hs + fif(is_outter, es,  0 ), 0},
        } 
    }

    for _, pos in ipairs(params.pos) do
        CreateRail{
            positionOffset = { pos, params.height },
            crossSectionShape = {
                {-hs + fif(is_outter, 0,  es), 0},
                { hs + fif(is_outter, 0, -es), 0}
            },
            perShapeNodeColorScalers = { 1, 1 },
            color = params.main_line.color,
            colorMode = params.main_line.color_mode,
            material = params.main_line.material
        }

        for _, shape in ipairs(edges_shapes) do
            CreateRail{
                positionOffset = { pos, params.height },
                crossSectionShape = shape,
                perShapeNodeColorScalers = { 1, 1 },
                color = params.border_line.color,
                colorMode = params.border_line.color_mode,
                material = params.border_line.material
            }
        end
    end
end

if is_wakeboard then
    -- left classic edge
    CreateRail{
        positionOffset = { -track_width - 0.2, 0 },
        crossSectionShape = {
            {-0.65,  0.5},
            { 0.30,  0.5},
            { 0.30, -3  },
            {-0.65, -3  }
        },
        perShapeNodeColorScalers = { 0.5, 1, 0, 0 },
        color = {255, 255, 255},
        colorMode = "highway",
        shader = "VertexColorUnlitTintedAddSmooth",
        layer = 13,
        renderqueue = 2750,
        stretch = 2
    }

    -- right classic edge
    CreateRail{
        positionOffset = {  track_width + 0.2, 0},
        crossSectionShape = {
            {-0.30,  0.5},
            { 0.65,  0.5},
            { 0.65, -3  },
            {-0.30, -3  }
        },
        perShapeNodeColorScalers = { 1, 0.5, 0, 0 },
        color = {255, 255, 255},
        colorMode = "highway",
        shader = "VertexColorUnlitTintedAddSmooth",
        layer = 13,
        renderqueue = 2750,
        stretch = 2
    }

    rails_create_border_aurora_lines{
        pos = skin_vars.shoulderlines,
        height = 0.25,
        main_line = {
            width = 0.25,
        },
        border_line = {
            width = 0.05,
            plasement = "outter",
        },
    }
else
    local aurora_edges_size = 0.2

    if is_shoulders_exist() then
        for i, line in ipairs(skin_vars.shoulderlines) do
            CreateRail{ 
                positionOffset = { line, 0.05 },
                crossSectionShape = {
                    {-.03, 0},
                    { .03, 0}
                },
                perShapeNodeColorScalers = { 1, 1 },
                color = {192, 192, 192},
                colorMode = "highway",
                shader = "VertexColorUnlitTinted",
                layer = 13,
                renderqueue = 2750,
                wrapnodeshape = false
            }
        end
    end

    rails_create_border_aurora_lines{
        pos = {-track_width + (aurora_edges_size / 2), track_width - (aurora_edges_size / 2)},
        height = 0,
        main_line = {
            width = aurora_edges_size,
        },
        border_line = {
            width = 0.03,
        },
    }
    
    CreateRail{
        positionOffset = { 0, 0 },
        crossSectionShape = {
            {-track_width + aurora_edges_size, 0},
            { track_width - aurora_edges_size, 0},
        },
        perShapeNodeColorScalers = { 1, 1 },
        color = {0, 0, 0, 0.75},
        colorMode = "static",
        shader = "VertexColorUnlitTintedAlpha",
        layer = 11,
        renderqueue = 3000
    }
end

do -- aurora itself
    local aurora_mtr, aurora_update = build_material_and_get_resources{
        shader = "VertexColorUnlitTintedAddDouble",
        shadercolors = {
            _Color = {127, 127, 127}
        },
        texture = "com_white.png",
        layer = 13,
        renderqueue = 2750,
        update_init = {
            shadercolors = {
                _Color = rgb:new {0, 0, 0}
            }
        }
    }

    if options.aurora.dynamic.enable then
        if options.aurora.dynamic.use_old_method then
            UpdateMaterial{
                material = aurora_mtr,
                shadercolors = {
                    _Color = {
                        colorsource = options.aurora.dynamic.color_source, 
                        scaletype = "intensity",
                        minscaler = options.aurora.dynamic.color_scaler.min, 
                        maxscaler = options.aurora.dynamic.color_scaler.max
                    }
                }
            }
        else
            material_onupdate_attach_colorscaler_handler{
                material_update_data = aurora_update,
                colorsource = options.aurora.dynamic.color_source, 
                scaletype = "intensity",
                minscaler = options.aurora.dynamic.color_scaler.min, 
                maxscaler = options.aurora.dynamic.color_scaler.max
            }
        end
    end

    local w = fif(is_wakeboard, 225, 150)
    local shapes = {
        {
            positionOffset = { -track_width, 0.0 },
            crossSectionShape = { {-w, 0}, { 0, 0}, { 0, 0}, {-w, 0} },
            perShapeNodeColorScalers = { 0, 1, 1, 0 },
        },
        {
            positionOffset = {  track_width, 0.0 },
            crossSectionShape = { { 0, 0}, { w, 0}, { w, 0}, { 0, 0} },
            perShapeNodeColorScalers = { 1, 0, 0, 1 },
        }
    }
    
    for _, shape in ipairs(shapes) do
        CreateRail{
            positionOffset = shape.positionOffset,
            crossSectionShape = shape.crossSectionShape,
            perShapeNodeColorScalers = shape.perShapeNodeColorScalers,
            color = {255, 255, 255},
            colorMode = "aurora",
            material = aurora_mtr,
            flatten = false,
            stretch = 2
        }
    end
end

-- print_all()
Update = onupdate
print("'"..SKIN.NAME.."' (v"..fif(IS_EXPERIMENTAL, SKIN.VERSION.."-dev", SKIN.VERSION)..") by "..SKIN.AUTHOR.."\n\n")
collectgarbage()